/****************************** Mqtt service **********************************
This file is part of the Ewings Esp8266 Stack.

This is free software. you can redistribute it and/or modify it but without any
warranty.

Author          : Suraj I.
created Date    : 1st June 2019
******************************************************************************/

#include <config/Config.h>

#if defined(ENABLE_MQTT_SERVICE)

#include "MqttServiceProvider.h"

/**
 * start mqtt service. initialize it with mqtt configs at database
 */
void MqttServiceProvider::begin( ESP8266WiFiClass* _wifi ){

  this->wifi = _wifi;
  this->_mqtt_payload = new char[ MQTT_PAYLOAD_BUF_SIZE ];
  memset( this->_mqtt_payload, 0, MQTT_PAYLOAD_BUF_SIZE );
  this->mqtt_client.mqttDataCallbackArgs = (uint32_t*)this;
  this->mqtt_client.OnData( MqttServiceProvider::handleMqttDataCb );

  this->handleMqttConfigChange();
}

/**
 * check and do mqtt publish on given configs.
 */
void MqttServiceProvider::handleMqttPublish(){

  #ifdef EW_SERIAL_LOG
  Logln( F("MQTT: handling mqtt publish interval") );
  #endif
  if( !this->mqtt_client.is_mqtt_connected() ) return;

  mqtt_pubsub_config_table _mqtt_pubsub_configs = __database_service.get_mqtt_pubsub_config_table();
  uint8_t mac[6];
  char macStr[18] = { 0 };
  wifi_get_macaddr(STATION_IF, mac);
  sprintf(macStr, "%02X:%02X:%02X:%02X:%02X:%02X", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

  for (uint8_t i = 0; i < MQTT_MAX_PUBLISH_TOPIC; i++) {

    __find_and_replace( _mqtt_pubsub_configs.publish_topics[i].topic, "[mac]", macStr, 2 );
    #ifdef EW_SERIAL_LOG
    Log( F("MQTT: publishing on topic : ") );Logln( _mqtt_pubsub_configs.publish_topics[i].topic );
    #endif
    if( strlen(_mqtt_pubsub_configs.publish_topics[i].topic) > 0 ){

      #ifdef ENABLE_MQTT_DEFAULT_PAYLOAD

        String _payload = "";

        #ifdef ENABLE_GPIO_SERVICE

          __gpio_service.appendGpioJsonPayload( _payload );
        #else

          _payload += "Hello from Esp Client : ";
          _payload += ESP.getChipId();
        #endif

        memset( this->_mqtt_payload, 0, MQTT_PAYLOAD_BUF_SIZE );
        _payload.toCharArray( this->_mqtt_payload, MQTT_PAYLOAD_BUF_SIZE );

      #else

      #endif

      if( this->_mqtt_publish_data_cb ) this->_mqtt_publish_data_cb( this->_mqtt_payload, MQTT_PAYLOAD_BUF_SIZE );
      __find_and_replace( this->_mqtt_payload, "[mac]", macStr, 2 );

      this->mqtt_client.Publish(
        _mqtt_pubsub_configs.publish_topics[i].topic,
        this->_mqtt_payload,
        strlen( this->_mqtt_payload ),
        _mqtt_pubsub_configs.publish_topics[i].qos < MQTT_MAX_QOS_LEVEL ?
        _mqtt_pubsub_configs.publish_topics[i].qos : MQTT_MAX_QOS_LEVEL,
        _mqtt_pubsub_configs.publish_topics[i].retain
      );
    }
  }
}

/**
 * handle mqtt subscribe cycle. suscribe to topics in pubsub configs
 */
void MqttServiceProvider::handleMqttSubScribe(){

  #ifdef EW_SERIAL_LOG
  Logln( F("MQTT: handling mqtt subscribe interval") );
  #endif
  if( !this->mqtt_client.is_mqtt_connected() ) return;

  mqtt_pubsub_config_table _mqtt_pubsub_configs = __database_service.get_mqtt_pubsub_config_table();

  for (uint8_t i = 0; i < MQTT_MAX_SUBSCRIBE_TOPIC; i++) {

    if( strlen(_mqtt_pubsub_configs.subscribe_topics[i].topic) > 0 && !this->mqtt_client.is_topic_subscribed(_mqtt_pubsub_configs.subscribe_topics[i].topic) ){

      this->mqtt_client.Subscribe(

        _mqtt_pubsub_configs.subscribe_topics[i].topic,
        _mqtt_pubsub_configs.subscribe_topics[i].qos < MQTT_MAX_QOS_LEVEL ?
        _mqtt_pubsub_configs.subscribe_topics[i].qos : MQTT_MAX_QOS_LEVEL
      );
    }
  }
}

/**
 * stop mqtt client
 *
 */
void MqttServiceProvider::stop(){

  this->mqtt_client.DeleteClient();
  __task_scheduler.clearInterval( this->_mqtt_timer_cb_id );
  this->_mqtt_timer_cb_id = 0;
}

/**
 * handle restart of mqtt services on config change from autherised client
 *
 * @param   int _mqtt_config_type
 */
void MqttServiceProvider::handleMqttConfigChange( int _mqtt_config_type ){


  mqtt_general_config_table _mqtt_general_configs = __database_service.get_mqtt_general_config_table();
  mqtt_pubsub_config_table _mqtt_pubsub_configs = __database_service.get_mqtt_pubsub_config_table();

  if( _mqtt_config_type == MQTT_GENERAL_CONFIG || _mqtt_config_type == MQTT_LWT_CONFIG ){

    this->mqtt_client.DeleteClient();
    int _stat = __task_scheduler.setTimeout( [&]() {

      mqtt_general_config_table _mqtt_general_configs = __database_service.get_mqtt_general_config_table();
      mqtt_lwt_config_table _mqtt_lwt_configs = __database_service.get_mqtt_lwt_config_table();

      uint8_t mac[6];
      char macStr[18] = { 0 };
      wifi_get_macaddr(STATION_IF, mac);
      sprintf(macStr, "%02X:%02X:%02X:%02X:%02X:%02X", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

      __find_and_replace( _mqtt_general_configs.username, "[mac]", macStr, 2 );
      __find_and_replace( _mqtt_general_configs.client_id, "[mac]", macStr, 2 );
      __find_and_replace( _mqtt_lwt_configs.will_message, "[mac]", macStr, 2 );

      if( this->mqtt_client.begin( this->wifi, &_mqtt_general_configs, &_mqtt_lwt_configs ) ){
        this->_mqtt_timer_cb_id = __task_scheduler.updateInterval(
          this->_mqtt_timer_cb_id,
          [&]() { this->mqtt_client.mqtt_timer(); },
          MILLISECOND_DURATION_1000
        );
      }else{
        __task_scheduler.clearInterval( this->_mqtt_timer_cb_id );
        this->_mqtt_timer_cb_id = 0;
      }
    }, MQTT_INITIALIZE_DURATION );
  }else if( _mqtt_config_type == MQTT_PUBSUB_CONFIG ){

    for ( uint16_t i = 0; i < this->mqtt_client.mqttClient.subscribed_topics.size(); i++) {

      bool _found = false;
      for (uint8_t j = 0; j < MQTT_MAX_SUBSCRIBE_TOPIC; j++) {

        if( __are_str_equals(
          _mqtt_pubsub_configs.subscribe_topics[j].topic,
          this->mqtt_client.mqttClient.subscribed_topics[i].topic,
          strlen( _mqtt_pubsub_configs.subscribe_topics[j].topic )
        ) ) _found = true;

      }
      if( !_found )
        this->mqtt_client.UnSubscribe( this->mqtt_client.mqttClient.subscribed_topics[i].topic );
    }

  }else{

  }

  if( _mqtt_pubsub_configs.publish_frequency > 0 ){
    this->_mqtt_publish_cb_id = __task_scheduler.updateInterval(
      this->_mqtt_publish_cb_id,
      [&]() { this->handleMqttPublish(); },
      _mqtt_pubsub_configs.publish_frequency*MILLISECOND_DURATION_1000
    );
  }else{
    __task_scheduler.clearInterval( this->_mqtt_publish_cb_id );
    this->_mqtt_publish_cb_id = 0;
  }

  if( _mqtt_general_configs.keepalive > 0 ){
    this->_mqtt_subscribe_cb_id = __task_scheduler.updateInterval(
      this->_mqtt_subscribe_cb_id,
      [&]() { this->handleMqttSubScribe(); },
      _mqtt_general_configs.keepalive*MILLISECOND_DURATION_1000
    );
  }else{
    __task_scheduler.clearInterval( this->_mqtt_subscribe_cb_id );
    this->_mqtt_subscribe_cb_id = 0;
  }

}

/**
 * set publish data callback. used to set data before send
 *
 * @param   MqttPublishDataCallback _cb
 */
void MqttServiceProvider::setMqttPublishDataCallback( MqttPublishDataCallback _cb ){

  this->_mqtt_publish_data_cb = _cb;
}

/**
 * set subscribe data callback. used to get data before received
 *
 * @param   MqttSubscribeDataCallback _cb
 */
void MqttServiceProvider::setMqttSubscribeDataCallback( MqttSubscribeDataCallback _cb ){

  this->_mqtt_subscribe_data_cb = _cb;
}

/**
 * handle subscribed topics data
 *
 */
void MqttServiceProvider::handleMqttDataCb( uint32_t *args, const char* topic, uint32_t topic_len, const char *data, uint32_t data_len ){

    if( __mqtt_service._mqtt_subscribe_data_cb ) __mqtt_service._mqtt_subscribe_data_cb(
      args, topic, topic_len, data, data_len
    );

    char *topicBuf = new char[topic_len+1], *dataBuf = new char[data_len+1];

    memcpy(topicBuf, topic, topic_len);
    topicBuf[topic_len] = 0;

    memcpy(dataBuf, data, data_len);
    dataBuf[data_len] = 0;

    // #ifdef EW_SERIAL_LOG
    // Logln(F("\n\nMQTT: service data callback"));
    // Serial.printf("MQTT: service Receive topic: %s, data: %s \n\n", topicBuf, dataBuf);
    // #endif

    #if defined( ENABLE_MQTT_DEFAULT_PAYLOAD ) && defined( ENABLE_GPIO_SERVICE )
    __gpio_service.applyGpioJsonPayload( dataBuf, data_len );
    #endif
    delete[] topicBuf; delete[] dataBuf;
}

#ifdef EW_SERIAL_LOG
/**
 * print mqtt configs
 */
void MqttServiceProvider::printMqttConfigLogs(){

  mqtt_general_config_table _mqtt_general_configs = __database_service.get_mqtt_general_config_table();
  mqtt_lwt_config_table _mqtt_lwt_configs = __database_service.get_mqtt_lwt_config_table();
  mqtt_pubsub_config_table _mqtt_pubsub_configs = __database_service.get_mqtt_pubsub_config_table();


  Logln(F("\nMqtt General Configs :"));
  Log( _mqtt_general_configs.host ); Log("\t");
  Log( _mqtt_general_configs.port ); Log("\t");
  Log( _mqtt_general_configs.security ); Log("\t");
  Log( _mqtt_general_configs.client_id ); Log("\t");
  Log( _mqtt_general_configs.username ); Log("\t");
  Log( _mqtt_general_configs.password ); Log("\t");
  Log( _mqtt_general_configs.keepalive ); Log("\t");
  Log( _mqtt_general_configs.clean_session ); Logln();

  Logln(F("\nMqtt Lwt Configs :"));
  Log( _mqtt_lwt_configs.will_topic ); Log("\t");
  Log( _mqtt_lwt_configs.will_message ); Log("\t");
  Log( _mqtt_lwt_configs.will_qos ); Log("\t");
  Log( _mqtt_lwt_configs.will_retain ); Logln();

  Logln(F("\nMqtt Pub Configs :"));
  for (uint8_t i = 0; i < MQTT_MAX_PUBLISH_TOPIC; i++) {

    if( strlen(_mqtt_pubsub_configs.publish_topics[i].topic) > 0 ){

      Log( _mqtt_pubsub_configs.publish_topics[i].topic ); Log("\t");
      Log( _mqtt_pubsub_configs.publish_topics[i].qos ); Log("\t");
      Log( _mqtt_pubsub_configs.publish_topics[i].retain ); Logln();
    }
  }

  Logln(F("\nMqtt Sub Configs :"));
  for (uint8_t i = 0; i < MQTT_MAX_SUBSCRIBE_TOPIC; i++) {

    if( strlen(_mqtt_pubsub_configs.subscribe_topics[i].topic) > 0 ){

      Log( _mqtt_pubsub_configs.subscribe_topics[i].topic ); Log("\t");
      Log( _mqtt_pubsub_configs.subscribe_topics[i].qos ); Logln();
    }
  }
}
#endif


MqttServiceProvider __mqtt_service;

#endif
